/*
 * Initialisation code for the stage 2 payload.
 *
 * Copyright 2009 (C) Qi Hardware Inc.
 * Author: Wolfgang Spraul <wolfgang@sharism.cc>
 *
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "sdram.h"

.text
.extern c_main
.extern _tlb_entry
.extern _exc_entry
.extern _irq_entry
.extern _end_entries
.extern _got_start
.extern _got_end
.extern _got_copy_start
.globl _start
.set noreorder

_start:
	b	real_start
	nop

	/* Apparently reserved region which, if used, breaks the USB Boot process. */

	.word	0x0
	.word	0x0
	.word	0x0
	.word	0x0
	.word	0x0
	.word	0x0
	.word	0x0
	.word	0x0

real_start:     
	/* Initialise the stack. */

	la $sp, 0x80080000

	/* Initialise the globals pointer. */

	lui $gp, %hi(_GLOBAL_OFFSET_TABLE_)
	ori $gp, $gp, %lo(_GLOBAL_OFFSET_TABLE_)

	move $k0, $ra

	/* Copy TLB handling instructions. */

	la $t0, _tlb_entry		/* start */
	li $t1, 0x80000000
	la $t2, _exc_entry		/* end */
	jal _copy
	nop

	/* Copy exception handling instructions. */

	la $t0, _exc_entry		/* start */
	li $t1, 0x80000180
	la $t2, _irq_entry		/* end */
	jal _copy
	nop

	/* Copy IRQ handling instructions. */

	la $t0, _irq_entry		/* start */
	li $t1, 0x80000200
	la $t2, _end_entries		/* end */
	jal _copy
	nop

	/* Copy the offset table for user mode programs. */

	la $t0, _got_start		/* start */
	la $t1, _got_copy_start
	la $t2, _got_end		/* end */
	li $t3, 0x80000000		/* adjustment */
	jal _copy_adjust
	nop

	move $ra, $k0

	/* Enable caching. */

	li $t0, CONFIG_CM_CACHABLE_NONCOHERENT
	mtc0 $t0, $16			/* CP0_CONFIG */
	nop

	/* Start the program. */

	la $t9, c_main			/* load the address of the routine */
	j c_main
	nop

_copy:
	/* Copy via $t3 the region from $t0 to $t2 into $t1. */

	lw $t3, 0($t0)
	addiu $t0, $t0, 4
	sw $t3, 0($t1)
	bne $t0, $t2, _copy
	addiu $t1, $t1, 4		/* executed in delay slot before branch */
	jr $ra
	nop

_copy_adjust:
	/* Copy via $t4 the region from $t0 to $t2 into $t1, adjusting by $t3. */

	lw $t4, 0($t0)
	addiu $t0, $t0, 4
	subu $t4, $t4, $t3
	sw $t4, 0($t1)
	bne $t0, $t2, _copy_adjust
	addiu $t1, $t1, 4		/* executed in delay slot before branch */
	jr $ra
	nop

.set reorder
